const config = require("config");
const Crypto = require("./Crypto");
const dayjs = require("dayjs");
const path = require("path");

class Order {
  static get STORAGE_KEY_PREFIX() {
    return "order_";
  }

  static get PURCHASE_TAX_AMOUNT() {
    return 0.07;
  }

  static get PURCHASE_STORAGE_TIME_MONTHS() {
    return 6;
  }
  // How many minutes is a user allowed to take for finishing the payment
  static get PURCHASE_TIME_LIMIT_MINUTES() {
    return 15;
  }

  static get PRICE_FOR_100() {
    return 2;
  }

  static get PRICE_FOR_250() {
    return 4;
  }

  // How many searches to wrap in a signed package
  static get PACKET_SIZE() {
    return 50;
  }

  /**
   * Data stored in Redis Database
   */
  #order_id;
  #expires_at;
  #amount;
  #unit_size;
  #price_per_unit;
  #encrypted_sales_receipts;
  #signatures;
  #payment_completed;
  #payment_method_link; // Stores a link to an entry of the payment methods payment i.e. PayPal order id

  /**
   * Data populated by context
   */
  #order_date;
  #order_path;
  #create_mode;
  #redis_client;

  constructor(
    order_id,
    expires_at,
    amount,
    unit_size,
    price_per_unit,
    encrypted_sales_receipts,
    signatures
  ) {
    this.#order_id = new String(order_id);
    this.#expires_at = dayjs(expires_at);
    this.#order_date = dayjs.unix(this.#order_id.substr(0, 10));

    this.#order_path = config.get("storage.data_path");
    if (process.env.NODE_ENV === "development") {
      this.#order_path = path.join(this.#order_path, "development");
    }
    this.#order_path = path.join(
      this.#order_path,
      this.#order_date.format("YYYY"),
      this.#order_date.format("MM")
    );

    this.#amount = parseInt(amount);
    this.#unit_size = parseInt(unit_size);
    this.#price_per_unit = parseFloat(price_per_unit);
    this.#encrypted_sales_receipts = encrypted_sales_receipts;

    if (signatures) {
      this.#signatures = signatures;
    } else {
      this.#signatures = [];
    }

    this.#payment_completed = false;
    this.#create_mode = true;

    let Redis = require("ioredis");
    this.#redis_client = new Redis({
      host: config.get("redis.host"),
    });
  }

  set_create_mode(create_mode) {
    this.#create_mode = create_mode;
  }

  getAmount() {
    return this.#amount;
  }

  getPricePerUnit() {
    return this.#price_per_unit;
  }

  getPaymentMethodLink() {
    return this.#payment_method_link;
  }

  getEncryptedSalesReceipts() {
    return this.#encrypted_sales_receipts;
  }

  getSignatures() {
    return this.#signatures;
  }

  isPaymentComplete() {
    return this.#payment_completed;
  }

  setPaymentMethodLink(payment_method_link) {
    this.#payment_method_link = payment_method_link;
  }

  async signOrder() {
    let mgcrypto = new Crypto();

    if (this.#signatures.length > 0) {
      return false;
    }

    let signed_encrypted_sales_receipts = mgcrypto.sign(
      this.#encrypted_sales_receipts,
      this.#order_date,
      this.#expires_at
    );
    this.#signatures = await signed_encrypted_sales_receipts;

    // Store amount of signed receipts for our records
    // So we know how much searches we have given out
    let fs = require("fs");
    let generated_file = path.join(this.#order_path, "generated.json");
    // Create directory if it does not exist
    if (!fs.existsSync(path.dirname(generated_file))) {
      fs.mkdirSync(path.dirname(generated_file), { recursive: true });
    }
    for (let i = 0; i < this.#signatures.length; i++) {
      fs.appendFileSync(generated_file, this.#order_id + `_${i}\n`);
    }

    return true;
  }

  static async LOAD_ORDER_FROM_ID(order_id) {
    return new Promise((resolve, reject) => {
      let Redis = require("ioredis");
      let redis_client = new Redis({
        host: config.get("redis.host"),
      });

      let redis_key = Order.STORAGE_KEY_PREFIX + order_id;

      let order_file = path.join(
        this.#order_path,
        this.#order_id.toString() + ".json"
      );
      redis_client.hgetall(redis_key).then((order_data) => {
        if (Object.keys(order_data).length === 0) {
          // Checking FS for order
          let fs = require("fs");
          if (fs.existsSync(order_file)) {
            order_data = JSON.parse(fs.readFileSync(order_file));
          } else {
            return reject("Could not find Order in our database! Checking FS");
          }
        }
        let loaded_order = new Order(
          order_data.order_id,
          order_data.expires_at,
          order_data.amount,
          order_data.unit_size,
          order_data.price_per_unit,
          JSON.parse(order_data.encrypted_sales_receipts),
          JSON.parse(order_data.signatures)
        );
        if (order_data.payment_method_link) {
          loaded_order.setPaymentMethodLink(
            JSON.parse(order_data.payment_method_link)
          );
        }
        loaded_order.set_create_mode(false);
        console.log(loaded_order.getEncryptedSalesReceipts());
        resolve(loaded_order);
      });
    });
  }

  async save() {
    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    let expiration = new dayjs();
    if (this.#create_mode) {
      expiration = expiration.add(Order.PURCHASE_TIME_LIMIT_MINUTES, "minute");
    } else {
      expiration = expiration.add(Order.PURCHASE_STORAGE_TIME_MONTHS, "month");
    }
    expiration = Math.round(expiration.diff() / 1000);

    let key_exists = await this.#redis_client.exists(redis_key);
    if (this.#create_mode && key_exists) {
      return Promise.reject("Cannot create a Order that already exists!");
    }
    let stored_data = {
      order_id: this.#order_id,
      expires_at: this.#expires_at.format("YYYY-MM-DD"),
      amount: this.#amount,
      unit_size: this.#unit_size,
      price_per_unit: this.#price_per_unit,
      encrypted_sales_receipts: JSON.stringify(this.#encrypted_sales_receipts),
      signatures: JSON.stringify(this.#signatures),
      payment_completed: this.#payment_completed,
      payment_method_link: JSON.stringify(this.#payment_method_link),
    };

    let storage_promise = this.#redis_client
      .hmset(redis_key, stored_data)
      .then((result) => {
        this.#create_mode = false;
      })
      .then((result) => {
        this.#redis_client.expire(redis_key, expiration);
      });

    await storage_promise;

    // If the Order payment is completed store the order additionally directly to harddisk
    let fs = require("fs");
    let order_file = path.join(
      this.#order_path,
      "orders",
      this.#order_id.toString() + ".json"
    );
    // Create directory if it does not exist
    if (!fs.existsSync(path.dirname(order_file))) {
      fs.mkdirSync(path.dirname(order_file), { recursive: true });
    }
    storage_promise = fs.writeFileSync(
      order_file,
      JSON.stringify(stored_data, null, 4)
    );

    await storage_promise;
  }

  async delete() {
    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    return this.#redis_client
      .del(redis_key)
      .then(() => {
        return true;
      })
      .catch(() => {
        return false;
      });
  }
}

module.exports = Order;
