var express = require("express");
var router = express.Router();

const config = require("config");
const Order = require("../../app/Order.js");

const CLIENT_ID = config.get(
  `payments.paypal.${process.env.NODE_ENV}.client_id`
);
const APP_SECRET = config.get(`payments.paypal.${process.env.NODE_ENV}.secret`);
const base = config.get(`payments.paypal.${process.env.NODE_ENV}.base`);

/* Client initiates payment */
router.post("/", async (req, res, next) => {
  // Order data is validated: Create and store the order in the redis database
  let order = new Order(
    req.body.order_id,
    req.body.expires_at,
    req.body.amount,
    req.body.unit_size,
    req.body.price_per_unit,
    req.body.encrypted_sales_receipts
  );

  order
    .save()
    .then(() => {
      // Order created on our side. Continue the payment with the selected provider
      createOrder(order).then((order_result) => res.json(order_result));
    })
    .catch((reason) => {
      return res.status(400).json({
        errors: [
          {
            msg: reason,
          },
        ],
      });
    });
});

// capture payment & store order information or fullfill order
router.post("/capture", async (req, res) => {
  Order.LOAD_ORDER_FROM_ID(req.body.order_id).then((loaded_order) => {
    loaded_order
      .signOrder()
      .then(() => {
        let paypal_order_id = loaded_order.getPaymentMethodLink().id;
        capturePayment(paypal_order_id)
          .then((captureData) => {
            loaded_order
              .save()
              .then(() => {
                captureData.signatures = loaded_order.getSignatures();
                captureData.order_id = req.body.order_id;
                captureData.expires_at = req.body.expires_at;
                res.json(captureData);
              })
              .catch((error) => {
                res.status(400).json({ errors: [{ msg: error }] });
              });
          })
          .catch((error) => {
            res.status(400).json({ errors: [{ msg: error }] });
          });
      })
      .catch((error) => {
        res.status(400).json({ errors: [{ msg: error }] });
      });

    //res.json(captureData);
    // TODO: store payment information such as the transaction ID
  });
});

module.exports = router;

//////////////////////

// PayPal API helpers

//////////////////////

// use the orders api to create an order

async function createOrder(loaded_order) {
  console.log(loaded_order);
  const accessToken = await generateAccessToken();

  const url = `${base}/v2/checkout/orders`;
  let tax_amount_per_unit = fixRounding(
    loaded_order.getPricePerUnit() * Order.PURCHASE_TAX_AMOUNT,
    2
  );
  let item_amount_per_unit = fixRounding(
    loaded_order.getPricePerUnit() - tax_amount_per_unit,
    2
  );

  let order = {
    intent: "CAPTURE",
    purchase_units: [
      {
        description: "MetaGer Pass Einkauf",
        amount: {
          currency_code: "EUR",
          value: fixRounding(
            item_amount_per_unit * loaded_order.getAmount() +
              tax_amount_per_unit * loaded_order.getAmount(),
            2
          ),
          breakdown: {
            item_total: {
              currency_code: "EUR",
              value: fixRounding(
                item_amount_per_unit * loaded_order.getAmount(),
                2
              ),
            },
            tax_total: {
              currency_code: "EUR",
              value: fixRounding(
                tax_amount_per_unit * loaded_order.getAmount(),
                2
              ),
            },
          },
        },
        items: [
          {
            name: "MetaGer Pass: Suchanfragen (250x)",
            quantity: loaded_order.getAmount(),
            unit_amount: {
              currency_code: "EUR",
              value: item_amount_per_unit,
            },
            tax: {
              currency_code: "EUR",
              value: tax_amount_per_unit,
            },
            category: "DIGITAL_GOODS",
            description:
              "MetaGer Pass Suchanfragen zur Nutzung der Suchmaschine MetaGer",
          },
        ],
      },
    ],
    application_context: {
      brand_name: "SUMA-EV",
      shipping_preference: "NO_SHIPPING",
    },
  };

  const response = await fetch(url, {
    method: "post",

    headers: {
      "Content-Type": "application/json",
      Authorization: `Bearer ${accessToken}`,
    },

    body: JSON.stringify(order),
  });

  const data = await response.json();
  loaded_order.setPaymentMethodLink({ id: data.id });
  return loaded_order.save().then(() => {
    return data;
  });
}

// use the orders api to capture payment for an order

async function capturePayment(orderId) {
  const accessToken = await generateAccessToken();

  const url = `${base}/v2/checkout/orders/${orderId}/capture`;

  const response = await fetch(url, {
    method: "post",

    headers: {
      "Content-Type": "application/json",

      Authorization: `Bearer ${accessToken}`,
    },
  });

  const data = await response.json();

  return data;
}

// generate an access token using client id and app secret

async function generateAccessToken() {
  const auth = Buffer.from(CLIENT_ID + ":" + APP_SECRET).toString("base64");

  const response = await fetch(`${base}/v1/oauth2/token`, {
    method: "post",

    body: "grant_type=client_credentials",

    headers: {
      Authorization: `Basic ${auth}`,
    },
  });

  const data = await response.json();

  return data.access_token;
}

/**
 *
 * Javascripts calculations with floating points produce weird artifacts. This method is to fix that issue
 *
 * @param {Float} value
 * @param {Int} precision
 * @returns
 */
function fixRounding(value, precision) {
  var power = Math.pow(10, precision || 0);
  return Math.round(value * power) / power;
}
